@@ -1,6 +1,6 @@
 (**************************************************************************)
 (*                                                                        *)
-(*    Copyright 2012-2015 OCamlPro                                        *)
+(*    Copyright 2012-2018 OCamlPro                                        *)
 (*    Copyright 2012 INRIA                                                *)
 (*                                                                        *)
 (*  All rights reserved. This file is distributed under the terms of the  *)
@@ -35,7 +35,7 @@ exception Nothing_to_do
 val source_pin:
   rw switch_state -> name ->
   ?version:version -> ?edit:bool -> ?opam:OpamFile.OPAM.t -> ?quiet:bool ->
-  ?force:bool -> ?ignore_extra_pins:bool ->
+  ?force:bool -> ?ignore_extra_pins:bool -> ?subpath: string -> ?locked:bool ->
   url option ->
   rw switch_state
 
@@ -43,6 +43,14 @@ val source_pin:
 val handle_pin_depends:
   rw switch_state -> package -> OpamFile.OPAM.t -> rw switch_state
 
+(** Fetch in parallel jobs a list of pins [name, url, subpath], and return the
+    successful ones.
+    Ask for confirmation to continue if a fetching fails.
+*)
+val fetch_all_pins:
+  'a switch_state -> ?working_dir:bool -> (name * url * string option) list ->
+  (url * string option) list
+
 (** Let the user edit a pinned package's opam file. If given, the version is put
     into the template in advance. Writes and returns the updated switch
     state. *)
@@ -57,6 +65,21 @@ val unpin_one: 'a switch_state -> package -> 'a switch_state
 (** List the pinned packages to the user. *)
 val list: 'a switch_state -> short:bool -> unit
 
+(** Scan pinning separator, used for printing and parsing by [scan] and
+    [parse_pins]. *)
+val scan_sep: char
+
+(** Scan for available packages to pin, and display it on stdout. If
+    [normalise] is true, displays it's normalised format
+    `name.version[scan_sep]url[scan_sep]subpath`. *)
+val scan: normalise:bool -> recurse:bool -> ?subpath: string -> url -> unit
+
+(** Detect if a string is a normalised format of [scan]. *)
+val looks_like_normalised: string list -> bool
+
+(** Parse the normalised form of [scan], and returns pinning informations. *)
+val parse_pins: string list -> (name * version option * url * string option) list
+
 (** Lints the given opam file, prints warnings or errors accordingly (unless
     [quiet]), upgrades it to current format, adds references to files below the
     'files/' subdir (unless the file is directly below the specified, local
